program mulserv;

{ A server that can handle multiple client connections. }

uses
   sockets, inetaux, myerror, linux;

const
   ListenPort : Word = $AFFE;
   MaxConn = 5;
   MaxClients = 5;

type
   ClientRec = record
      cSock : LongInt;
      adstr : String;
      sin, sout : Text;
   end;

var
   lSock, uSock : LongInt;
   sAddr : TInetSockAddr;
   Len, i, j : LongInt;
   Line : String;
   Clients : array[1..MaxClients] of ClientRec;
   NumClients : LongInt;
   MaxFD : LongInt;
   ReadSet : FDSet;
   sin, sout : Text;

begin
   lSock := Socket(af_inet, sock_stream, 0);
   if lSock = -1 then SockError('Socket: ');
   
   with sAddr do
   begin
      Family := af_inet;
      Port := htons(ListenPort);
      Addr := 0;
   end;
   
   if not Bind(lSock, sAddr, sizeof(sAddr)) then SockError('Bind: ');
   if not Listen(lSock, MaxConn) then SockError('Listen: ');
   
   Say('Waiting...');
   
   fd_zero(ReadSet);
   NumClients := 0;
   
   repeat
      MaxFD := 0;
      for i := 1 to NumClients do with Clients[i] do
      begin
         fd_set(cSock, ReadSet);
         if cSock > MaxFD then MaxFD := cSock;
      end;
      fd_set(lSock, ReadSet);
      if lSock > MaxFD then MaxFD := lSock;
      Inc(MaxFD);
      
      Select(MaxFD, @ReadSet, nil, nil, nil);  { No timeout! }
      
      { New connections? }
      
      if fd_isset(lSock, ReadSet) then
      begin
         Say('Incoming connection.');
         Len := sizeof(sAddr);
         uSock := Accept(lSock, sAddr, Len);
         if uSock = -1 then SockSay('Accept: ')
         else
         begin
            if NumClients < MaxClients then
            begin
               Inc(NumClients);
               with Clients[NumClients] do
               begin
                  cSock := uSock;
                  Sock2Text(cSock, sin, sout);
                  Reset(sin);
                  Rewrite(sout);
                  adstr := AddrToStr(sAddr.Addr);
                  Say('Accepted connection from ' + adstr);
               end;
            end
            else  { client limit reached }
            begin
               Sock2Text(uSock, sin, sout);
               Rewrite(sout);
               Writeln(sout, 'Sorry, we are fully booked.');
               Close(sout);
               Shutdown(uSock, 2);
            end;
         end;
      end;
      
      { And/or an event on an existing connection? }
      
      for i := 1 to NumClients do if fd_isset(Clients[i].cSock, ReadSet) then
      begin
         if eof(Clients[i].sin) then  { Connection has been closed? }
         begin
            with Clients[i] do
            begin
               Close(sin);
               Close(sout);
               Shutdown(cSock, 2);
               Say('Disconnected ' + adstr);
            end;
            for j := i to NumClients - 1 do Clients[j] := Clients[j + 1];
            Dec(NumClients);
         end
         else  { No -> Data can be read }
         begin
            Say('Received message from ' + Clients[i].adstr + '...');
            Readln(Clients[i].sin, Line);
            Say(Line);
            for j := 1 to NumClients do if i <> j then
               Writeln(Clients[j].sout, Line);
         end;
      end;
   until False;
end.
